/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Programmer notes:

--the code described in the article
	http://www.ibm.com/developerworks/library/j-benchmark2/index.html#dsat
may be found in the file
	N08_Part2_DataStructureAccess.original
in this directory.

--THE CURRENT CODE THAT YOU NOW SEE BELOW IS A MAJOR CHANGE FROM THE CODE ORIGINALLY USED IN THE ARTICLE:
	--the command line arguments have been totally revised:
		--there are many more
		--they are now expressed as proper switches (i.e. name/value pairs of the form -name value)
	--the data structure access pattern is now specified by the new Mode inner enum
	--the new TaskAbstract class replaces the old TaskRunnableAbstract class
	--the integers array is no longer a field of TaskAbstract, since it does not need to be retained for the non-array data structures
		--instead, most of the data structures simply use the result of makeIntegers in their construcctor and then can garbage collect it
	--KEY ALGORITHM CHANGE: I realized that I could use the values of the data structure itself to specify the access pattern (sequential versus random)
		--this allowed me to eliminate:
			--the random versions of the task classes, reducing code bloat
			--the LFSR code (inside the old RandomIndex inner class) that I used to use to generate random indices,
				which eliminates that as a source of benchmarking overhead
			--the additional accessing of integers inside the Map classes (on top of the call to get)
				which eliminates that as a source of benchmarking overhead
			--the XOR computation that I used to do on the old state field
				which eliminates that as a source of benchmarking overhead
		--make sure and study the TaskAbstract inner class to understand the new algorithm
	--added many more List and Map data structure test classes, since the general pattern is now so easy to code for using the new ListAccess and MapAccess base classes
	--this class now only does one benchmark per invocation of main
		--the script file must generate all the scenarios to run
		--MADE THIS CHANGE BECAUSE GET BEST RESULTS WHEN THERE IS ONLY A SINGLE BENCHMARK PER JVM SESSION
	--the benchmark result from main is now appended to 2 different files instead of the console
		--one file has the full result, the other a mean-only summary (this file is convenient for pasting into a spreadsheet)

+++ may need to run with certain JVM options:
	--the 1.6.0_14 release notes
		http://java.sun.com/javase/6/webnotes/6u14.html
	say:
		Improvement TreeMap Iteration
		6u14 includes an experimental implementation of java.util.TreeMap that can improve the performance of applications that iterate over TreeMaps very frequently. This implementation is used when running with the -XX:+AggressiveOpts option.

+++ future enhancements: any more "array-like" data structures whose access time I should benchmark?
	--EnumMap (but would need to autogenerate its class file on the fly to handle whatever size is...)

+++ definitely should download and try benchmarking google's collections:
	http://code.google.com/p/google-collections/
*/

import bb.io.FileUtil;
import bb.util.ArrayUtil;
import bb.util.Benchmark;
import bb.util.Check;
import bb.util.DateUtil;
import bb.util.Properties2;
import java.io.File;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.IdentityHashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;
import java.util.WeakHashMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentSkipListMap;
import java.util.concurrent.CopyOnWriteArrayList;

/**
* For the article
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html">Robust Java benchmarking, Part 2: Statistics And solutions</a>,
* contains the code relevant to
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#fig3">Figure 3</a>,
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#fig4">Figure 4</a>,
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#fig5">Figure 5</a>,
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#listing2">Listing 2</a>,
* and <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#listing3">Listing 3</a>.
* <p>
* @author Brent Boyer
*/
public class N08_Part2_DataStructureAccess {
	
	// -------------------- switch constants --------------------
	
	protected static final String class_key = "-class";
	
	protected static final String mode_key = "-mode";
	
	protected static final String size_key = "-size";
	
	protected static final String pathPrefix_key = "-pathPrefix";
	
	/** Specifies all the switch keys which can legally appear as command line arguments to {@link #main main}. */
	protected static final List<String> keysLegal = Arrays.asList(
		class_key, mode_key, size_key, pathPrefix_key
	);
	
	// -------------------- main and helper methods --------------------
	
	/**
	* Entry point to the functionality offered by this class.
	* <p>
	* Results on 2008-03-27 on my 2.2 GHz Intel Core 2 Duo E4500 machine (server jvm) are in the
	* <code>../results/N08_Part2_DataStructureAccess_Desktop_XXX.txt</code> series of files.
	* <p>
	* Results on 2008-03-27 on my 1.167 GHz SPARC-Enterprise-T5220 machine (server jvm) are in the
	* <code>../results/N08_Part2_DataStructureAccess_N2_XXX.txt</code> series of files.
	* Note: the N2 results are CPU times, not elapsed times (since the tests are all single threaded and Solaris supports this so well)
	* <p>
	* The files above contain the full Benchmark reports.
	* Edited versions of the relevant data (i.e. just the means) of all of the above results,
	* plus all of the associated Figures in the article, are found in the spreadsheet
	* <code>../results/N08_Part2_DataStructureAccess.xls</code>.
	* <p>
	* Results on 2009-12-20 on my 2.5 GHz Intel Xeon E5420 machine (server jvm) are in the
	* <code>.../results/newPostArticleResults/dataStructureAccess directory.
	* <p>
	* @param args must be non-null and have exactly 8 elements;
	* these 8 elements must consist of these 4 command line switches (name/value pairs);
	* the switches may be in any order:
	* -class [the name of one of the task inner classes e.g. ArrayAccess],
	* -mode [either sequential or random],
	* -size [the number of elements in the data structure, as an int],
	* -pathPrefix [valid file system path prefix; the final part of the file name will be generated by this class]
	* @throws Exception (or some subclass) if any problem happens
	*/
	public static void main(String[] args) throws Exception {
			// uncomment these lines only if need to generate a geometric series of sizes for a script file:
		//int[] sizesGeometric = makeGeometricSeries(2, 1 * 1024 * 1024, 30);
		//System.out.println("Here is makeGeometricSeries(2, 1 * 1024 * 1024, 50): " + StringUtil.toString(sizesGeometric, " "));
		//if (true) return;
		
		Check.arg().notEmpty(args);
		
		long t1 = System.nanoTime();
		System.out.println(DateUtil.getTimeStamp() + " benchmark started");
		
		Properties2 switches = new Properties2(args).checkKeys(keysLegal);
		
		String className = switches.getProperty(class_key);
		Mode mode = Mode.parse( switches.getProperty(mode_key) );
		int size =  switches.getInt(size_key);
		String pathPrefix = switches.getProperty(pathPrefix_key);
		
		TaskAbstract task = getTask(className, size, mode);
		
		Benchmark.Params params = new Benchmark.Params(true, size);	// true means measure many executions; supply the data structure size for the number of actions per execution (since each element is accessed once)
		//params.setNumberMeasurements(3);	// only use this when testing in orer to get results quickly
		//params.setMeasureCpuTime(true);	// default is false: measure elapsed time
		
		Benchmark benchmark = new Benchmark(task, params);
		
		task.postBenchmarkCheck();
		
		resultsToFile(className, size, mode, pathPrefix, benchmark, task.toString());
		
		long t2 = System.nanoTime();
		System.out.println(DateUtil.getTimeStamp() + " benchmark finished; execution time = " + ((t2 - t1) * 1e-9) + " seconds");
	}
	
	@SuppressWarnings("unchecked")
	protected static TaskAbstract getTask(String className, int size, Mode mode) throws Exception {
		String nameParent = N08_Part2_DataStructureAccess.class.getName();
		Class c = Class.forName(nameParent + "$" + className);
		Constructor cons = c.getDeclaredConstructor(int.class, Mode.class);
		return (TaskAbstract) cons.newInstance(size, mode);
	}
	
	protected static void resultsToFile(String className, int size, Mode mode, String pathPrefix, Benchmark benchmark, String result) throws Exception {
		outputMeanOnly(className, mode, size, pathPrefix, benchmark, result);
		outputFull(className, mode, size, pathPrefix, benchmark, result);
	}
	
	/** Writes the mean-only results to a file that is suitable for loading into a spreadsheet for graphing. */
	protected static void outputMeanOnly(String className, Mode mode, int size, String pathPrefix, Benchmark benchmark, String result) throws Exception {
		Benchmark.Stats stats = benchmark.getStatsAction();
		double mean = stats.getMean();
		double deltaLower = stats.getMeanLower() - mean;
		double deltaUpper = stats.getMeanUpper() - mean;
		String meanInfo = className + "\t" + mode + "\t" + size + "\t" + mean + "\t" + deltaLower + "\t" + deltaUpper + "\t" + result + "\n";
		
		File file = new File(pathPrefix + "_meanOnly.txt");
		
		if (!file.exists()) {
			String header ="className" + "\t" + "mode" + "\t" + "size" + "\t" + "mean (seconds)" + "\t" + "deltaLower" + "\t" + "deltaUpper" + "\t" + "result" + "\n";
			FileUtil.writeString( header, file, true );
		}
		
		FileUtil.writeString( meanInfo, file, true );	// CRITICAL: supply true here to append to the file
		System.out.println("Mean-only results written to: " + file.getPath());
	}
	
	/** Writes the full results to a file. */
	protected static void outputFull(String className, Mode mode, int size, String pathPrefix, Benchmark benchmark, String result) throws Exception {
		String fullInfo = "\n" + className + ", size = " + size + ", mode = " + mode + ": " + benchmark.toStringFull() + "\n" + "result = " + result + "\n";
		
		File file = new File(pathPrefix + "_full.txt");
		
		FileUtil.writeString( fullInfo, file, true );	// CRITICAL: supply true here to append to the file
		System.out.println("Full results written to: " + file.getPath());
	}
	
	/**
	* Generates the double-valued geometric series <code>[start, start * r, start * r^2, ..., start * r^(n-1)]</code>.
	* In other words, generates n values, the first of which is start, and the last of which is end,
	* which have the property that the ratio of sucessive elements is a constant r.
	* This last property causes a logarithmic plot of the values to have uniform point density.
	* <p>
	* The result that is actually returned is an int rounded version of that double-valued sequence
	* that is guaranteed to have no duplicate elements.
	* Thus, the size of the result may be less than n (but will never exceed it)
	* and the ratio of successive elements may not exactly equal a constant r (because of the rounding).
	* <p>
	* @param start the first element in the result, guaranteed
	* @param end the last element in the result, guaranteed
	* @param n number of elements that would like the result to contain; the actual size of the result may be less than this
	*/
	protected static int[] makeGeometricSeries(int start, int end, int n) throws IllegalArgumentException {
		if (n < 2) throw new IllegalArgumentException("n = " + n + " is < 2");
		double ratio = ((double) end) / start;
		if (ratio < 0) throw new IllegalStateException("ratio = end/start = " + start + "/" + end + " = " + ratio + " is < 0");
		
		SortedSet<Integer> set = new TreeSet<Integer>();	// CRITICAL: use a SortedSet, and NOT a List, because in the loop below sometimes when d and r are both near 1, val does not change between loop iterations because of the rounding, so want to avoid duplicates
		double r = Math.pow( ratio, 1.0/(n - 1) );
		for (double d = start; d <= end; d *= r) {	// d is the values in double form, which need to keep track of for accuracy purposes
			if (Double.isNaN(d)) throw new IllegalStateException("d is NaN; occurred for start = " + start + ", end = " + end + ", n = " + n);
			
			int val = (int) Math.round(d);	// val is the int rounded version of d and is what will actually be returned
			set.add(val);
		}
		set.add(end);	// CRITICAL: must do this because sometimes, due to rounding error in the d *= r computation, the loop above does not include this
		
		int[] result = new int[set.size()];
		int i = 0;
		for (Integer integer : set) result[i++] = integer;
		if (result[0] != start) throw new IllegalStateException("result[0] = " + result[0] + " != start = " + start);
		if (result[result.length - 1] != end) throw new IllegalStateException("result[result.length - 1] = " + result[result.length - 1] + " != end = " + end);
		if (result.length > n) throw new IllegalStateException("result.length = " + result.length + " > n = " + n);
		return result;
	}
	
	// -------------------- Mode (static inner enum) --------------------
	
	/**
	* Enum of all the supported data structure access pattern modes.
	* <p>
	* This enum is multithread safe: it is stateless (except for the enumeration of values, which are immutable).
	* <p>
	* Like all java enums, this enum is Comparable and Serializable.
	* <p>
	* @see <a href="http://java.sun.com/j2se/1.5.0/docs/guide/language/enums.html">Enum documentation</a>
	*/
	protected static enum Mode {
		
		sequential,
		random;
		
		protected static Mode parse(String s) throws IllegalArgumentException {
			if ("sequential".equalsIgnoreCase(s)) return sequential;
			else if ("random".equalsIgnoreCase(s)) return random;
			else throw new IllegalArgumentException("s = " + s + " does not parse into a known Mode");
		}
		
	}
	
	// -------------------- task base classes and actual implementations (static inner classes) --------------------
	
	/**
	* Abstract base class for Tasks which measure the access time of "array-like" data structures:
	* arrays, array-like Lists (e.g. ArrayList and Vector), and associative arrays (i.e. Maps).
	* <p>
	* <b>Subclasses must obey all the rules below.</b>
	* <p>
	* <i>Rule #1:</i> a dedicated subclass must written for each type of data structure.
	* <p>
	* <i>Rule #2:</i> each subclass must implement its <code>run</code> method like:
	* <pre><code>
		public void run() {
			for (int i = 0; i < {@link #size}; i++) {
				{@link #index} = [the value of the element at the current value of index in the data structure];
			}
		}
	* </code></pre>
	* <p>
	* <i>Rule #3:</i> all subclass data structures must be isomorphic to the result of {@link #makeIntegers makeIntegers},
	* namely, their (index/key, value) pairs must be identical to <code>makeIntegers</code>.
	* This requirement is automatically satisfied by {@link ArrayAccess} and its subclasses.
	* Those subclasses which are Lists should clear themselves and then add all the elements of <code>makeIntegers</code>.
	* Those subclasses which are Maps should clear themselves
	* and then in a loop over the elements of <code>makeIntegers</code> do <code>map.put( integer, integer );</code>.
	* <p>
	* <i>The reason for these rules is that the values of <code>makeIntegers</code> are cleverly choosen
	* so that the loop in <code>run</code> accesses the data structure's elements
	* in a specific pattern (either sequential or random).</i>
	* <p>
	* To see how this works for the sequential case,
	* suppose that the data structure is the result of <code>makeIntegers</code>, so that the code is
	* <pre><code>
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				index = integers[index.intValue()];
			}
		}
	* </code></pre>
	* (This is how {@link ArrayAccess} is implemented.)
	* If all (but the last) elements of <code>integers</code> satisfy <code>integers[i] equals i + 1</code>,
	* then the loop in <code>run</code> accesses the elements of <code>integers</code> sequentially
	* (assuming that index is initialized to 0, and that the last element of <code>integers</code> also is 0).
	* <p>
	* The other functionality defined in this base class is DCE prevention, given that subclasses will be Runnable task classes.
	*/
	protected abstract static class TaskAbstract implements Runnable {
		
		/** Access pattern mode of the data structure. */
		protected final Mode mode;
		
		/** Size of the data structure. */
		protected final int size;
		
		/**
		* Index/key of the next element of the data structure that will be accessed.
		* <p>
		* Additionally, since the instances that will be benchmarked are Runnables,
		* this field stores the value of some computation done in {@link #run run}
		* <a href="http://www.ibm.com/developerworks/java/library/j-benchmark1.html#dce">to prevent DCE</a>.
		*/
		protected Integer index;	// Note: even tho this is not a primitive, it will not cause performance issues when other code below autounboxes it (the reverse is what causes problems)
		
		/**
		* Constructor.
		* <p>
		* @param size specifies the number of elements in the data structure
		* @param mode specifies the data structure access pattern:
		* @throws IllegalArgumentException if size < 1; mode is null
		*/
		protected TaskAbstract(int size, Mode mode) throws IllegalArgumentException {
			Check.arg().positive(size);
			Check.arg().notNull(mode);
			
			this.size = size;
			this.mode = mode;
		}
		
		/**
		* Returns an Integer[] that always contains the numbers [0, size - 1].
		* The order of the result's elements are chosen so that if it is used in a run loop as described in this class's javadocs,
		* then its elements will be accessed in the order specified by {@link #mode}.
		* <p>
		* Side effect: assigns {@link #index} to the 0th element of the result.
		*/
		protected Integer[] makeIntegers() {
			Integer[] indices = makeIndices();
			index = indices[0];
			
			Integer[] integers = new Integer[size];
			for (int i = 0; i < size - 1; i++) {	// CRITICAL: < size - 1 NOT size because there is a special case (see below)
				integers[ indices[i] ] = indices[i + 1];
			}
			integers[ indices[size - 1] ] = indices[0];	// special case: final element accessed needs to point to the first accessed element, completing the cycle; this also causes the same sequence of values of index to be computed each time run is called
			
			checkIntegers(integers, indices);
			
			return integers;
		}
		
		/**
		* Returns an Integer[] that specifies the sequence of indices of the relevant data structure
		* that will be accessed by the loop in {kink #run run}.
		* For example, if the result equals {2, 1, 3, 0},
		* then means that access element 2, then element 1, then element 3, then element 0 IN THAT ORDER.
		*/
		protected Integer[] makeIndices() {
			Integer[] indices = new Integer[size];
			switch (mode) {
				case sequential:
					for (int i = 0; i < size; i++) indices[i] = i;	// i.e. the run loop should access the elements of indices in their natural order
					break;
				case random:
					for (int i = 0; i < size; i++) indices[i] = i;
					Random r = new Random(0);	// CRITICAL: must use a Random with the same seed (0 here) to guarantee consistency across runs
					ArrayUtil.shuffle(indices, r);	// causes the run loop to access the data structure elements in random order
					break;
				default:
					throw new IllegalStateException("mode = " + mode + " is currently unsupported");
			}
			return indices;
		}
		
		/**
		* Checks that integers is suitable for use in the run method described in the class javadocs
		* (i.e. every element is accessed in an iteration identical to the run loop in the order specified by indices).
		*/
		protected void checkIntegers(Integer[] integers, Integer[] indices) throws IllegalStateException {
			int[] indicesVisited = new int[size];
			int ii = indices[0];
			for (int i = 0; i < size; i++) {
				indicesVisited[i] = ii;
				ii = integers[ii];	// same step as is used in the run loop
			}
			for (int i = 0; i < size; i++) {
				if (indicesVisited[i] != indices[i]) throw new IllegalStateException("indicesVisited[" + i + "] = " + indicesVisited[i] + " != indices[" + i + "] = " + indices[i]);
			}
		}
		
		/**
		* Returns the String representation of {@link #index}
		* <a href="http://www.ibm.com/developerworks/java/library/j-benchmark1.html#dce">to prevent DCE</a>.
		*/
		@Override public String toString() {
			return String.valueOf(index);
		}
		
		protected void postBenchmarkCheck() throws IllegalStateException {
			if (index == null) throw new IllegalStateException("index == null");
		}
		
	}
	
	// ========== array related ==========
	
	/** Task class for benchmarking the access time of an array ({@link #integers}). */
	protected static class ArrayAccess extends TaskAbstract {
		protected final Integer[] integers;
		
		protected ArrayAccess(int size, Mode mode) throws IllegalArgumentException {
			super(size, mode);
			
			integers = makeIntegers();
		}
		
		@Override public void run() {
			for (int i = 0; i < integers.length; i++) {
				index = integers[index.intValue()];	// see TaskAbstract javadocs for explanation
			}
		}
	}
	
	/** Same code as {@link ArrayAccess} except synchronize on {@link #integers} around each access. */
	protected static class ArrayAccessSynchronized extends ArrayAccess {
		protected ArrayAccessSynchronized(int size, Mode mode) throws IllegalArgumentException {
			super(size, mode);
		}
		
		@Override public void run() {
			for (int i = 0; i < integers.length; i++) {
				synchronized(integers) {
					index = integers[index.intValue()];	// see TaskAbstract javadocs for explanation
				}
			}
		}
	}
	
	// ========== List related ==========
	
	/**
	* Same code as {@link ArrayAccess} except use a List that is equivalent to the result of {@link #makeIntegers makeIntegers}.
	* This class is used as the base class for all List-type classes.
	*/
	protected abstract static class ListAccess extends TaskAbstract {
		protected final List<Integer> list;
		
		protected ListAccess(int size, Mode mode, List<Integer> list) throws IllegalArgumentException {
			super(size, mode);
			
			list.clear();	// done for safety...
			//for (Integer integer : makeIntegers()) { list.add(integer); }
				// cannot do the simple line above, because at least one subclass (CopyOnWriteArrayListAccess) becomes REALLY slow when size is large due to having to copy the entire structure for each add, so must do a bulk addAll
			list.addAll( Arrays.asList( makeIntegers() ) );
			this.list = list;
//this.list = Collections.unmodifiableList(list);	// use this line instead if want to see its effect
		}
		
		@Override public void run() {
			for (int i = 0; i < size; i++) {
				index = list.get(index.intValue());	// see TaskAbstract javadocs for explanation
			}
		}
	}
	
	/** Simply causes its superclass {@link ListAccess} to use an ArrayList. */
	protected static class ArrayListAccess extends ListAccess {
		protected ArrayListAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new ArrayList<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link ListAccess} to use a Vector. */
	protected static class VectorAccess extends ListAccess {
		protected VectorAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new Vector<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link ListAccess} to use a CopyOnWriteArrayList. */
	protected static class CopyOnWriteArrayListAccess extends ListAccess {
		protected CopyOnWriteArrayListAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new CopyOnWriteArrayList<Integer>() );
		}
	}
	
	// ========== Map related ==========
	
	/**
	* Same code as {@link ArrayAccess} except use a Map that is equivalent to the result of {@link #makeIntegers makeIntegers}.
	* This class is used as the base class for all Map-type classes.
	*/
	protected abstract static class MapAccess extends TaskAbstract {
		protected final Map<Integer,Integer> map;
		
		protected MapAccess(int size, Mode mode, Map<Integer,Integer> map) throws IllegalArgumentException {
			super(size, mode);
			
			map.clear();	// done for safety...
			Integer[] integers = makeIntegers();
			//for (int i = 0; i < size; i++) { map.put(i, integers[i]); }
				// cannot do the simple line above, because at least two subclasses (IdentityHashMapAccess and WeakHashMapAccess) will not give proper results unless you ensure that all the keys and values are elements of integers
			Map<Integer,Integer> tmp = new HashMap<Integer,Integer>();
			for (int i = 0; i < size; i++) { tmp.put(integers[i], integers[i]); }
			for (int i = 0; i < size; i++) { map.put(tmp.get(i), integers[i]); }	// this is what ensures that the keys of map are elements of integers
			this.map = map;
//this.map = Collections.unmodifiableMap(map);	// use this line instead if want to see its effect
		}
		
		@Override public void run() {
			for (int i = 0; i < size; i++) {
				index = map.get(index);	// see TaskAbstract javadocs for explanation
			}
		}
	}
	
	/** Simply causes its superclass {@link MapAccess} to use a HashMap. */
	protected static class HashMapAccess extends MapAccess {
		protected HashMapAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new HashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapAccess} to use a Hashtable. */
	protected static class HashtableAccess extends MapAccess {
		protected HashtableAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new Hashtable<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapAccess} to use an IdentityHashMap. */
	protected static class IdentityHashMapAccess extends MapAccess {
		protected IdentityHashMapAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new IdentityHashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapAccess} to use a LinkedHashMap. */
	protected static class LinkedHashMapAccess extends MapAccess {
		protected LinkedHashMapAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new LinkedHashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapAccess} to use a TreeMap. */
	protected static class TreeMapAccess extends MapAccess {
		protected TreeMapAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new TreeMap<Integer,Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link MapAccess} to use a WeakHashMap. */
	protected static class WeakHashMapAccess extends MapAccess {
		protected WeakHashMapAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new WeakHashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	// ========== ConcurrentMap related ==========
	
	/**
	* Simply causes its superclass {@link MapAccess} to use a ConcurrentHashMap.
	* This is a best case for ConcurrentHashMap access, since it specifies concurrencyLevel = 1.
	*/
	protected static class ConcurrentHashMapAccess1 extends MapAccess {
		protected ConcurrentHashMapAccess1(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new ConcurrentHashMap<Integer,Integer>(size, 0.75f, 1) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/**
	* Simply causes its superclass {@link MapAccess} to use a ConcurrentHashMap.
	* This is a more realistic case for ConcurrentHashMap access, since it uses the default concurrencyLevel = 16.
	*/
	protected static class ConcurrentHashMapAccess16 extends MapAccess {
		protected ConcurrentHashMapAccess16(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new ConcurrentHashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapAccess} to use a ConcurrentSkipListMap. */
	protected static class ConcurrentSkipListMapAccess extends MapAccess {
		protected ConcurrentSkipListMapAccess(int size, Mode mode) throws IllegalArgumentException {
			super( size, mode, new ConcurrentSkipListMap<Integer,Integer>() );
		}
	}
	
	// ========== ThreadLocalAccess related ==========
	
	/** Similar code to {@link MapAccess} except use a ThreadLocal. */
	@SuppressWarnings("unchecked")
	protected static class ThreadLocalAccess extends TaskAbstract {
		protected final ThreadLocal<Integer>[] integersPerThread;
		
		protected ThreadLocalAccess(int size, Mode mode) throws IllegalArgumentException {
			super(size, mode);
			
			final Integer[] integers = makeIntegers();
			
			integersPerThread = new ThreadLocal[size];
			for (int i = 0; i < size; i++) {
				final Integer integer = integers[i];
				integersPerThread[i] = new ThreadLocal<Integer>() {
					protected Integer initialValue() { return integer; }
				};
			}
		}
		
		@Override public void run() {
			for (int i = 0; i < size; i++) {
				index = integersPerThread[index.intValue()].get();	// see TaskAbstract javadocs for explanation
			}
		}
	}
	
}
