/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Programmer notes:

--THIS CODE IS EXTRA: IT WAS NOT PART OF THE ORIGINAL ARTICLE
	http://www.ibm.com/developerworks/library/j-benchmark2/index.html#dsat
(altho it probably ought to have been...)

+++ future enhancements: there are more data structures whose access time I should benchmark
	--EnumSet/EnumMap (but would need to autogenerate their class files on the fly to handle whatever m is...)

+++ definitely should download and try benchmarking google's collections:
	http://code.google.com/p/google-collections/
*/

import bb.util.Benchmark;
import bb.util.Check;
import bb.util.DateUtil;
import bb.util.Properties2;
import java.lang.reflect.Constructor;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.IdentityHashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;
import java.util.WeakHashMap;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ConcurrentSkipListMap;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.PriorityBlockingQueue;

/**
* Benchmarks the time to completely iterate thru all the elements of common data structures.
* <p>
* This class is distinct from executing {@link N08_Part2_DataStructureAccess} in sequential access mode:
* instead of making individual (element by element) access requests of a data structure,
* it always does a complete iteration thru the data structure, usually using the data structure's Iterator.
* <p>
* One consequence of using the data structure's Iterator is that many classes can deliver much higher performance
* when tested here compared to N08_Part2_DataStructureAccess.
* A prime example is TreeMap: it has quite poor performance when tested under N08_Part2_DataStructureAccess,
* even in sequential access mode, but it does much better when tested by this class.
* <p>
* Another consequence is that this class is able to benchmark vastly more data structures
* than N08_Part2_DataStructureAccess (which is limited to array-like data structures).
* For instance, this class can benchmark Sets, Queues, and Dequeues.
* <p>
* @author Brent Boyer
*/
public class N14_Extra_DataStructureIteration {
	
	// -------------------- switch constants --------------------
	
	protected static final String class_key = "-class";
	
	protected static final String size_key = "-size";
	
	protected static final String pathPrefix_key = "-pathPrefix";
	
	/** Specifies all the switch keys which can legally appear as command line arguments to {@link #main main}. */
	protected static final List<String> keysLegal = Arrays.asList(
		class_key, size_key, pathPrefix_key
	);
	
	// -------------------- main and helper methods --------------------
	
	/**
	* Entry point to the functionality offered by this class.
	* <p>
	* Results on 2009-05-06 on my 2.5 GHz Intel Xeon E5420 machine (server jvm) are in the
	* <code>.../results/newPostArticleResults/dataStructureIteration directory.
	* <p>
	* @param args must be non-null and have exactly 6 elements;
	* these 6 elements must consist of these 4 command line switches (name/value pairs);
	* the switches may be in any order:
	* -class [the name of one of the task inner classes e.g. ArrayAccess],
	* -size [the number of elements in the data structure, as an int],
	* -pathPrefix [valid file system path prefix; the final part of the file name will be generated by this class]
	* @throws Exception (or some subclass) if any problem happens
	*/
	public static void main(String[] args) throws Exception {
		Check.arg().notEmpty(args);
		
		long t1 = System.nanoTime();
		System.out.println(DateUtil.getTimeStamp() + " benchmark started");
		
		Properties2 switches = new Properties2(args).checkKeys(keysLegal);
		
		String className = switches.getProperty(class_key);
		int size =  switches.getInt(size_key);
		String pathPrefix = switches.getProperty(pathPrefix_key);
		
		TaskAbstract task = getTask(className, size);
		
		Benchmark.Params params = new Benchmark.Params(true, size);	// true means measure many executions; supply the data structure size for the number of actions per execution (since each element is accessed once)
		//params.setNumberMeasurements(3);	// only use this when testing in orer to get results quickly
		//params.setMeasureCpuTime(true);	// default is false: measure elapsed time
		
		Benchmark benchmark = new Benchmark(task, params);
		
		N08_Part2_DataStructureAccess.resultsToFile(className, size, null, pathPrefix, benchmark, task.toString());
		
		long t2 = System.nanoTime();
		System.out.println(DateUtil.getTimeStamp() + " benchmark finished; execution time = " + ((t2 - t1) * 1e-9) + " seconds");
	}
	
	@SuppressWarnings("unchecked")
	protected static TaskAbstract getTask(String className, int size) throws Exception {
		String nameParent = N14_Extra_DataStructureIteration.class.getName();
		Class c = Class.forName(nameParent + "$" + className);
		Constructor cons = c.getDeclaredConstructor(int.class);
		return (TaskAbstract) cons.newInstance(size);
	}
	
	// -------------------- task base classes and actual implementations (static inner classes) --------------------
	
	/**
	* Abstract base class for Tasks which measure the iteration time of Iterable data structures:
	* arrays, Collections, Maps.
	* <p>
	* <b>Subclasses must obey all the rules below.</b>
	* <p>
	* <i>Rule #1:</i> a dedicated subclass must written for each type of data structure.
	* <p>
	* <i>Rule #2:</i> each subclass must implement its <code>run</code> method something like:
	* <pre><code>
		public void run() {
			dce = 0;	// do this so that each call of this method should
			for (Integer i : [some Iterable]) {
				{@link #dce} ^= i;
			}
		}
	* </code></pre>
	* <p>
	* <i>Rule #3:</i> all subclass data structures must be isomorphic to the result of {@link #makeIntegers makeIntegers},
	* namely, their (index/key, value) pairs must be identical to <code>makeIntegers</code>.
	* This requirement is atuomatically satisfied by {link ArrayIteration} and its subclasses.
	* Those subclasses which are Collections should clear themselves and then add all the elements of <code>makeIntegers</code>.
	* Those subclasses which are Maps should clear themselves
	* and then in a loop over the elements of <code>makeIntegers</code> do <code>map.put( integer, integer );</code>.
	* <p>
	* The reason for these rules is to ensure that each data structure is as similar as possible to make comparison easier.
	* <p>
	* The other functionality defined in this base class is DCE prevention, given that subclasses will be Runnable task classes.
	*/
	protected abstract static class TaskAbstract implements Runnable {
		
		/** Size of the data structure. */
		protected final int size;
		
		/** Used <a href="http://www.ibm.com/developerworks/java/library/j-benchmark1.html#dce">to prevent DCE</a>. */
		protected int dce;
		
		protected TaskAbstract(int size) throws IllegalArgumentException {
			Check.arg().positive(size);
			
			this.size = size;
		}
		
		/** Returns an Integer[] that always contains the numbers [0, size - 1] in order. */
		protected Integer[] makeIntegers() {
			Integer[] array = new Integer[size];
			for (int i = 0; i < size; i++) {
				array[i] = i;
			}
			return array;
		}
		
		/**
		* Returns the String representation of {@link #dce}
		* <a href="http://www.ibm.com/developerworks/java/library/j-benchmark1.html#dce">to prevent DCE</a>.
		*/
		@Override public String toString() {
			return String.valueOf(dce);
		}
		
	}
	
	// ========== array related ==========
	
	/** Task class for benchmarking the access time of an array of Integers. */
	protected static class ArrayIteration extends TaskAbstract {
		protected final Integer[] integers;
		
		protected ArrayIteration(int size) throws IllegalArgumentException {
			super(size);
			
			integers = makeIntegers();
		}
		
		@Override public void run() {
			dce = 0;	// do this so that each call of this method should result in the same value of dce at the end
			for (Integer i : integers) {
				dce ^= i;	// see TaskAbstract javadocs for explanation
			}
		}
	}
	
	/** Same code as {@link ArrayIteration} except synchronize on {@link #integers} around each access. */
	protected static class ArrayIterationSynchronized extends ArrayIteration {
		protected ArrayIterationSynchronized(int size) throws IllegalArgumentException {
			super(size);
		}
		
		@Override public void run() {
			synchronized(integers) {	// NOTE: unlike N08_Part2_DataStructureAccess, here we put the synchronized on the outside of the for loop
				dce = 0;	// do this so that each call of this method should result in the same value of dce at the end
				for (Integer i : integers) {
					dce ^= i;	// see TaskAbstract javadocs for explanation
				}
			}
		}
	}
	
	// ========== Collection related ==========
	
	/**
	* Same code as {@link ArrayIteration} except use a Collection that is equivalent to {@link ArrayIteration#integers}.
	* This class is used as the base class for most Collection-type classes.
	*/
	protected abstract static class CollectionIteration extends TaskAbstract {
		protected final Collection<Integer> collection;
		
		protected CollectionIteration(int size, Collection<Integer> collection) throws IllegalArgumentException {
			super(size);
			
			collection.clear();	// done for safety...
			//for (Integer integer : makeIntegers()) { collection.add(integer); }
				// cannot do the simple line above, because at least one subclass (CopyOnWriteArrayListIteration) becomes REALLY slow when size is large due to having to copy the entire structure for each add, so must do a bulk addAll
			collection.addAll( Arrays.asList( makeIntegers() ) );
			this.collection = collection;
//this.collection = Collections.unmodifiableCollection(collection);	// use this line instead if want to see its effect
		}
		
		@Override public void run() {
			dce = 0;	// do this so that each call of this method should result in the same value of dce at the end
			for (Integer i : collection) {
				dce ^= i;	// see TaskAbstract javadocs for explanation
			}
		}
	}
	
		// +++++ List related +++++
	
	/** Simply causes its superclass {@link CollectionIteration} to use an ArrayList. */
	protected static class ArrayListIteration extends CollectionIteration {
		protected ArrayListIteration(int size) throws IllegalArgumentException {
			super( size, new ArrayList<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a LinkedList. */
	protected static class LinkedListIteration extends CollectionIteration {
		protected LinkedListIteration(int size) throws IllegalArgumentException {
			super( size, new LinkedList<Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a Stack. */
	protected static class StackIteration extends CollectionIteration {
		protected StackIteration(int size) throws IllegalArgumentException {
			super( size, new Stack<Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a Vector. */
	protected static class VectorIteration extends CollectionIteration {
		protected VectorIteration(int size) throws IllegalArgumentException {
			super( size, new Vector<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a CopyOnWriteArrayList. */
	protected static class CopyOnWriteArrayListIteration extends CollectionIteration {
		protected CopyOnWriteArrayListIteration(int size) throws IllegalArgumentException {
			super( size, new CopyOnWriteArrayList<Integer>() );
		}
	}
	
		// +++++ Set related +++++
	
	/** Simply causes its superclass {@link CollectionIteration} to use a HashSet. */
	protected static class HashSetIteration extends CollectionIteration {
		protected HashSetIteration(int size) throws IllegalArgumentException {
			super( size, new HashSet<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a LinkedHashSet. */
	protected static class LinkedHashSetIteration extends CollectionIteration {
		protected LinkedHashSetIteration(int size) throws IllegalArgumentException {
			super( size, new LinkedHashSet<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a TreeSet. */
	protected static class TreeSetIteration extends CollectionIteration {
		protected TreeSetIteration(int size) throws IllegalArgumentException {
			super( size, new TreeSet<Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a ConcurrentSkipListSet. */
	protected static class ConcurrentSkipListSetIteration extends CollectionIteration {
		protected ConcurrentSkipListSetIteration(int size) throws IllegalArgumentException {
			super( size, new ConcurrentSkipListSet<Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a CopyOnWriteArraySet. */
	protected static class CopyOnWriteArraySetIteration extends CollectionIteration {
		protected CopyOnWriteArraySetIteration(int size) throws IllegalArgumentException {
			super( size, new CopyOnWriteArraySet<Integer>() );
		}
	}
	
		// +++++ Queue related +++++
	
	/** Simply causes its superclass {@link CollectionIteration} to use an ArrayBlockingQueue. */
	protected static class ArrayBlockingQueueIteration extends CollectionIteration {
		protected ArrayBlockingQueueIteration(int size) throws IllegalArgumentException {
			super( size, new ArrayBlockingQueue<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	//DelayQueue skip: don't want to deal with the Delayed type...
	
	/** Simply causes its superclass {@link CollectionIteration} to use a ConcurrentLinkedQueue. */
	protected static class ConcurrentLinkedQueueIteration extends CollectionIteration {
		protected ConcurrentLinkedQueueIteration(int size) throws IllegalArgumentException {
			super( size, new ConcurrentLinkedQueue<Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a LinkedBlockingQueue. */
	protected static class LinkedBlockingQueueIteration extends CollectionIteration {
		protected LinkedBlockingQueueIteration(int size) throws IllegalArgumentException {
			super( size, new LinkedBlockingQueue<Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a PriorityBlockingQueue. */
	protected static class PriorityBlockingQueueIteration extends CollectionIteration {
		protected PriorityBlockingQueueIteration(int size) throws IllegalArgumentException {
			super( size, new PriorityBlockingQueue<Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a PriorityQueue. */
	protected static class PriorityQueueIteration extends CollectionIteration {
		protected PriorityQueueIteration(int size) throws IllegalArgumentException {
			super( size, new PriorityQueue<Integer>() );
		}
	}
	
		// +++++ Deque related +++++
	
	/** Simply causes its superclass {@link CollectionIteration} to use an ArrayDeque. */
	protected static class ArrayDequeIteration extends CollectionIteration {
		protected ArrayDequeIteration(int size) throws IllegalArgumentException {
			super( size, new ArrayDeque<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link CollectionIteration} to use a LinkedBlockingDeque. */
	protected static class LinkedBlockingDequeIteration extends CollectionIteration {
		protected LinkedBlockingDequeIteration(int size) throws IllegalArgumentException {
			super( size, new LinkedBlockingDeque<Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	// ========== Map related ==========
	
	/**
	* Same code as {@link ArrayIteration} except use a Map that is equivalent to the result of {@link #makeIntegers makeIntegers}.
	* This class is used as the base class for all Map-type classes.
	*/
	protected abstract static class MapIteration extends TaskAbstract {
		protected final Map<Integer,Integer> map;
		
		protected MapIteration(int size, Map<Integer,Integer> map) throws IllegalArgumentException {
			super(size);
			
			map.clear();	// done for safety...
			Integer[] integers = makeIntegers();
			//for (int i = 0; i < size; i++) { map.put(i, integers[i]); }
				// cannot do the simple line above, because at least one subclass (WeakHashMapIteration) will not give proper results unless you ensure that all the keys and values are elements of integers
			Map<Integer,Integer> tmp = new HashMap<Integer,Integer>();
			for (int i = 0; i < size; i++) { tmp.put(integers[i], integers[i]); }
			for (int i = 0; i < size; i++) { map.put(tmp.get(i), integers[i]); }	// this is what ensures that the keys of map are elements of integers
			this.map = map;
//this.map = Collections.unmodifiableMap(map);	// use this line instead if want to see its effect
		}
		
		@Override public void run() {
			dce = 0;	// do this so that each call of this method should result in the same value of dce at the end
			for (Map.Entry<Integer,Integer> entry : map.entrySet()) {
				dce ^= entry.getValue();	// see TaskAbstract javadocs for explanation; could call entry.getKey() just as readily?
			}
		}
	}
	
	/** Simply causes its superclass {@link MapIteration} to use a HashMap. */
	protected static class HashMapIteration extends MapIteration {
		protected HashMapIteration(int size) throws IllegalArgumentException {
			super( size, new HashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapIteration} to use a Hashtable. */
	protected static class HashtableIteration extends MapIteration {
		protected HashtableIteration(int size) throws IllegalArgumentException {
			super( size, new Hashtable<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapIteration} to use an IdentityHashMap. */
	protected static class IdentityHashMapIteration extends MapIteration {
		protected IdentityHashMapIteration(int size) throws IllegalArgumentException {
			super( size, new IdentityHashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapIteration} to use a LinkedHashMap. */
	protected static class LinkedHashMapIteration extends MapIteration {
		protected LinkedHashMapIteration(int size) throws IllegalArgumentException {
			super( size, new LinkedHashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapIteration} to use a TreeMap. */
	protected static class TreeMapIteration extends MapIteration {
		protected TreeMapIteration(int size) throws IllegalArgumentException {
			super( size, new TreeMap<Integer,Integer>() );
		}
	}
	
	/** Simply causes its superclass {@link MapIteration} to use a WeakHashMap. */
	protected static class WeakHashMapIteration extends MapIteration {
		protected WeakHashMapIteration(int size) throws IllegalArgumentException {
			super( size, new WeakHashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	// ========== ConcurrentMap related ==========
	
	/**
	* Simply causes its superclass {@link MapIteration} to use a ConcurrentHashMap.
	* This is a best case for ConcurrentHashMap access, since it specifies concurrencyLevel = 1.
	*/
	protected static class ConcurrentHashMapIteration1 extends MapIteration {
		protected ConcurrentHashMapIteration1(int size) throws IllegalArgumentException {
			super( size, new ConcurrentHashMap<Integer,Integer>(size, 0.75f, 1) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/**
	* Simply causes its superclass {@link MapIteration} to use a ConcurrentHashMap.
	* This is a more realistic case for ConcurrentHashMap access, since it uses the default concurrencyLevel = 16.
	*/
	protected static class ConcurrentHashMapIteration16 extends MapIteration {
		protected ConcurrentHashMapIteration16(int size) throws IllegalArgumentException {
			super( size, new ConcurrentHashMap<Integer,Integer>(size) );	// CRITICAL: must optimally size it, else could blow out the cache prematurely due to wasted array sizing
		}
	}
	
	/** Simply causes its superclass {@link MapIteration} to use a ConcurrentSkipListMap. */
	protected static class ConcurrentSkipListMapIteration extends MapIteration {
		protected ConcurrentSkipListMapIteration(int size) throws IllegalArgumentException {
			super( size, new ConcurrentSkipListMap<Integer,Integer>() );
		}
	}
	
}
