/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Programmer notes:

--THIS CODE IS EXTRA: IT WAS NOT PART OF THE ORIGINAL ARTICLE
	http://www.ibm.com/developerworks/library/j-benchmark2/index.html#dsat
(altho I wish it was, as it would have been a great example...)

--background:
	--I was asked by GETCO during an interview to write pure Java code which estimates the sizes of the data caches on a CPU
	--I solved this problem as follows: benchmark the average random access time of an int[] as a function of the array size,
	for many different sizes
	--plotting the result should reveal the sizes of the various caches on the CPU
	(they should be the points where the access time suddenly rises)
	--this class is the benchmark code that I used

--an int[] is used because it is trivial to estimate the total amount of memory used by a given array:
just multiply the array size by the 4 bytes used by each Java int
	--in contrast, the size of objects like Integers can vary across JVMs, hardware, etc

--RANDOM (as opposed to sequential) array access must be used,
because you do not want some sophisticated memory prefetcher or something to distort the results
*/

import bb.util.ArrayUtil;
import bb.util.Benchmark;
import bb.util.Check;
import bb.util.DateUtil;
import bb.util.Properties2;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

/**
* Benchmarks the random access time of an int[] of a specified size.
* <p>
* @author Brent Boyer
*/
public class N15_Extra_ArrayRandomAccessVersusSize {
	
	// -------------------- switch constants --------------------
	
	protected static final String size_key = "-size";
	
	protected static final String pathPrefix_key = "-pathPrefix";
	
	/** Specifies all the switch keys which can legally appear as command line arguments to {@link #main main}. */
	protected static final List<String> keysLegal = Arrays.asList(
		size_key, pathPrefix_key
	);
	
	// -------------------- main and helper methods --------------------
	
	/**
	* Entry point to the functionality offered by this class.
	* <p>
	* Results on 2009-04-05 on my 2.5 GHz Intel Xeon E5420 machine (server jvm) are in the
	* <code>.../results/newPostArticleResults/cacheSizes directory.
	* <p>
	* @param args must be non-null and have exactly 4 elements;
	* these 4 elements must consist of these 2 command line switches (name/value pairs);
	* the switches may be in any order:
	* -size [the number of elements in the data structure, as an int],
	* -pathPrefix [valid file system path prefix; the final part of the file name will be generated by this class]
	* @throws Exception (or some subclass) if any problem happens
	*/
	public static void main(String[] args) throws Exception {
			// uncomment these lines only if need to generate a geometric series of sizes for a script file:
		//int[] sizesGeometric = N08_Part2_DataStructureAccess.makeGeometricSeries(2, 16 * 1024 * 1024, 200);
		//System.out.println("Here is makeGeometricSeries(2, 16 * 1024 * 1024, 200): " + StringUtil.toString(sizesGeometric, " "));
		//if (true) return;
		
		Check.arg().notEmpty(args);
		
		long t1 = System.nanoTime();
		System.out.println(DateUtil.getTimeStamp() + " benchmark started");
		
		Properties2 switches = new Properties2(args).checkKeys(keysLegal);
		
		int size =  switches.getInt(size_key);
		String pathPrefix = switches.getProperty(pathPrefix_key);
		
		Runnable task = new TaskIntArray(size);
		
		Benchmark.Params params = new Benchmark.Params(true, size);	// true means measure many executions; supply the data structure size for the number of actions per execution (since each element is accessed once)
		//params.setNumberMeasurements(3);	// only use this when testing in orer to get results quickly
		//params.setMeasureCpuTime(true);	// default is false: measure elapsed time
		
		Benchmark benchmark = new Benchmark(task, params);
		
		N08_Part2_DataStructureAccess.resultsToFile("TaskIntArray", size, null, pathPrefix, benchmark, task.toString());
		
		long t2 = System.nanoTime();
		System.out.println(DateUtil.getTimeStamp() + " benchmark finished; execution time = " + ((t2 - t1) * 1e-9) + " seconds");
	}
	
	// -------------------- TaskIntArray (static inner class) --------------------
	
	/**
	* Identical to a combination of {@link N08_Part2_DataStructureAccess.TaskAbstract} and {@link N08_Part2_DataStructureAccess.ArrayAccess}
	* except that is customized for random access of an int[] instead of arbitrary access of an Integer[] (the Programmer notes above explain why).
	* All further javadocs have been stripped from the code below to eliminate duplication,
	* so consult TaskAbstract and ArrayAccess if you need explanation.
	*/
	protected static class TaskIntArray implements Runnable {
		
		protected final int[] array;
		
		protected int index;
		
		protected TaskIntArray(int size) throws IllegalArgumentException {
			Check.arg().positive(size);
			
			array = makeArray(size);
		}

		protected int[] makeArray(int size) {
			int[] indices = makeIndices(size);
			index = indices[0];
			
			int[] ints = new int[size];
			for (int i = 0; i < size - 1; i++) {	// CRITICAL: < size - 1 NOT size because there is a special case (see below)
				ints[ indices[i] ] = indices[i + 1];
			}
			ints[ indices[size - 1] ] = indices[0];	// special case: final element accessed needs to point to the first accessed element, completing the cycle; this also causes the same sequence of values of index to be computed each time run is called
			
			checkArray(ints, indices);
			
			return ints;
		}
		
		protected int[] makeIndices(int size) {
			int[] indices = new int[size];
			for (int i = 0; i < size; i++) indices[i] = i;
			Random r = new Random(0);	// CRITICAL: must use a Random with the same seed (0 here) to guarantee consistency across runs
			ArrayUtil.shuffle(indices, r);	// causes the run loop to access the data structure elements in random order
			return indices;
		}
		
		protected void checkArray(int[] ints, int[] indices) throws IllegalStateException {
			int[] indicesVisited = new int[ints.length];
			int ii = indices[0];
			for (int i = 0; i < ints.length; i++) {
				indicesVisited[i] = ii;
				ii = ints[ii];	// same step as is used in the run loop
			}
			for (int i = 0; i < ints.length; i++) {
				if (indicesVisited[i] != indices[i]) throw new IllegalStateException("indicesVisited[" + i + "] = " + indicesVisited[i] + " != indices[" + i + "] = " + indices[i]);
			}
		}
		
		@Override public void run() {
			for (int i = 0; i < array.length; i++) {
				index = array[index];
			}
		}
		
		@Override public String toString() {
			return String.valueOf(index);
		}
		
	}
	
}
